//
// game.cpp
//
// Copyright (C) David Joffe 1995-2001
//
// License: GNU GPL
//
// 1995/07/28
/*--------------------------------------------------------------------------*/

#include "mission.h"
#include "hero.h"
#include "inventory.h"
#include "thing.h"
#include "graph.h"
#include "game.h"
#include "graph.h"
#include "block.h"
#include "level.h"
#include "djgamelib.h"
#include "menu.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#ifdef WIN32
#else
// gack, posix
#include <unistd.h>
#endif


// sound files

char *g_szSoundFiles[SOUND_MAX] =
{
	"data/sounds/pickup.wav",
	"data/sounds/laser.wav",
	"data/sounds/bark.wav",
	"data/sounds/wooeep.wav"
};
SOUND_HANDLE g_iSounds[SOUND_MAX];


// positions of various elements on screen

#define SCORE_X (28*8)
#define SCORE_Y (3*8)

#define HEALTH_X (28*8)
#define HEALTH_Y (7*8)

#define LIVES_X (288)
#define LIVES_Y (96)


// some game constants

// Yeah its pretty low but thats what I was aiming for back in the EGA days
// and I don't feel like changing it right now. I might still though.
#define FRAME_RATE (18.0f)

#define LIVES_MAX 6
#define LIVES_INITIAL 3
#define HEALTH_INITIAL 10

#define LEVCHAR_FOREA(x,y) (*( g_pLevel + (y) * 512 + (x) * 4 + 0 ))
#define LEVCHAR_FOREB(x,y) (*( g_pLevel + (y) * 512 + (x) * 4 + 1 ))
#define LEVCHAR_BACKA(x,y) (*( g_pLevel + (y) * 512 + (x) * 4 + 2 ))
#define LEVCHAR_BACKB(x,y) (*( g_pLevel + (y) * 512 + (x) * 4 + 3 ))

#define CHECK_SOLID(a,b)  ( GET_EXTRA( (a), (b), EXTRA_FLAGS ) & FLAG_SOLID )

int g_iLives = LIVES_INITIAL;
unsigned char *g_pLevel = NULL;
int ilevel = 0;
bool bShowDebugInfo = false;
vector<CThing *> g_apThings;

int g_nFirepower = 1;



#define BULLET_WIDTH 16
#define BULLET_HEIGHT 8
class CBullet
{
public:
	CBullet() { x=0; y=0; dx=0; dy=0; nAnim=0; bMovedOnce=false; bDrawnOnce=false; }
	void Tick()
	{
		if (bMovedOnce)
		{
			x += dx;
			y += dy;
			nAnim = (nAnim+1)%4;
		}
		bMovedOnce = true;
	}
	void Draw()
	{
		djgDrawImageAlpha(pVisView,
			g_pCurMission->GetSpriteData(5)->m_pImage,
			bDrawnOnce ? ((nAnim)%16)*16 : (dx<0 ? 4 : 5)*16,
			((0)/16)*16,
			WORLDX2VIEW(x),
			WORLDY2VIEW(y)-4,
			16,
			16);
		bDrawnOnce = true;
	}
	int x; // World (pixel) X coordinate
	int y; // World (pixel) Y coordinate
	int dx;
	int dy;
	bool bMovedOnce;
	bool bDrawnOnce;
	int nAnim;
};
vector<CBullet*> g_apBullets;

// Bullet system
int  ifirepower = 3;
void DrawBullets();


vector<float> afTimeTaken;
#define MAX_DEBUGGRAPH 128


int iTeleportCountdown = 0;  // non-zero means countdown to teleport
int teleport_x;
int teleport_y;

char    *FILE_GAMESKIN = "data/gameskin.tga";
djImage *pSkinGame=NULL; // Main game view skin (while playing)
djImage *pBackground = NULL; // Level background image

/*--------------------------------------------------------------------------*/
struct SMenuItem gameMenuItems[] =
{
   { false, "{~~~~~~~~~~~~~~~~~}" },
   { true,  "|  Continue       |" },
   { true,  "|  Save Game      |" },
   { true,  "|  Restore Game   |" },
   { true,  "|  Abort Game     |" },
   { false, "[~~~~~~~~~~~~~~~~~]" },
   { false, NULL }     
};
unsigned char gameMenuCursor[] = { 128, 129, 130, 131, 0 };
CMenu gameMenu;
/*--------------------------------------------------------------------------*/

//animation
int anim4_count;
int foocount;

// hero
int x = 64, y = 50;      // x,y = hero's absolute positioning in level
int xo = 60, yo = 45;    // xo,yo = top-left corner of view for scrolling
int hero_dir = 1;        // hero direction, left==0, right==1
int hero_picoffs = 0;    // hero animation image index offset
int x_small=0, xo_small=0; // x_small == 0 ? hero at x : hero at x + 8 pixels
// xo_small == 0 ? view offset on x boundary : view offset 8 pixels

// when hurt, set to positive "framecount" that
// counts down. Can't get hurt again until it reaches 0.
int nHurtCounter = 0;

const int MODE_NORMAL = 0;
const int MODE_JUMPING = 1;

//game-play
unsigned int score = 0;
int  health;
bool g_bGameRunning = false;
int  key_up    = 0;
int  key_down  = 0;
int  key_left  = 0;
int  key_right = 0;
int  key_ctrl  = 0;
int  key_shoot = 0;
int  hero_mode = MODE_NORMAL;

// jumping
struct SJumpInfo
{
	int size;         // number of offsets
	int * jump_diffs; // offsets (y axis)
};
int iJumpNormal[] = { -1, -1, -1,  0,  0, 0,  1,  1,  1 };
int iJumpBoots[]  = { -1, -1, -1, -1,  0,  0, 0,  1,  1,  1,  1 };
struct SJumpInfo jumpNormal = {  9, iJumpNormal }; // "Normal" jump offsets
struct SJumpInfo jumpBoots  = { 11, iJumpBoots  }; // Jump offsets with boots
struct SJumpInfo * pJumpInfo; // Points to current jump info, normal or boots
int jump_pos = 0; // Offset into the "jump info" array of y-axis offsets

void HeroSetJumpMode(EJump eJump)
{
	switch (eJump)
	{
	case JUMP_NORMAL:     pJumpInfo = &jumpNormal; break;
	case JUMP_POWERBOOTS: pJumpInfo = &jumpBoots; break;
	}
}

/*-----------------------------------------------------------*/
// Game functions
/*-----------------------------------------------------------*/
// Once off initialization stuff
void GameInitialSetup()
{
	printf("GameInitialSetup()\n" );
	// Set up the in-game menu
	gameMenu.m_clrBack = djColor(0,0,129);
	gameMenu.m_iSize = 0;
	gameMenu.m_items = gameMenuItems;
	gameMenu.m_szCursor = gameMenuCursor;
	gameMenu.m_xOffset = -1; // Calculate menu position for us
	gameMenu.m_yOffset = -1;

	printf("GameInitialSetup(): loading sounds\n" );
	// Load the game sounds
	int i;
	for ( i=0; i<SOUND_MAX; i++ )
	{
		g_iSounds[i] = djSoundLoad( g_szSoundFiles[i] );
	}

	// If main game skin not yet loaded, load it
	if (pSkinGame==NULL)
	{
		if (NULL != (pSkinGame = new djImage))
			pSkinGame->Load( FILE_GAMESKIN );
		else
			djMSG( "GameDrawSkin(): error loading skin image file.\n" );
	}
}

// Per-game initialization
void GameSetup()
{
	// fixme
	g_iLives  = LIVES_INITIAL; // Initial number of lives
	health    = HEALTH_INITIAL; // Initial health
FIXME_IfHeroInMidAirThenPossibleStuffUp: // is it?
	HeroSetJumpMode(JUMP_NORMAL);

	ilevel = 0;

	score = 0;

	g_nFirepower = 1;

	djMSG( "GameSetup(): Loading sprites\n" );
	if (0 != GameLoadSprites())
	{
		djMSG( "GameSetup(): error loading sprites.\n" );
		return;
	}

}

// Per-level initialization
void GameLevelSetup()
{
	int i;

	djMSG( "GameLevelSetup()\n" );


	// (1) Initialize all relevant variables for entering a new level
	iTeleportCountdown = 0;

	// Key states variables
	key_up    = 0;
	key_down  = 0;
	key_left  = 0;
	key_right = 0;
	key_ctrl  = 0;
	key_shoot = 0;

	health = HEALTH_INITIAL;
	
	hero_mode = MODE_NORMAL; // Standing around
	nHurtCounter = 0; // Not currently hurting
	anim4_count=0; // animation count 0
	hero_picoffs=0;
	x_small = 0; // hero not half-block offset
	xo_small = 0; // view not half-block offset
	// just in case level doesn't contain a starting block ..
	xo = 0;
	yo = 0;
	relocate_hero( 20, 20 );
	hero_dir = 1;
	
	// clear list of "things"
	for ( i=0; i<(int)g_apThings.size(); i++ )
	{
		delete g_apThings[i];
	}
	g_apThings.clear();
	
	// (2) Load the currently selected level
	char * szfilename = g_pCurMission->GetLevel( ilevel )->GetFilename( );
	
	// always keep current level loaded at slot 0
	TRACE("GameLevelSetup(): level_load( %s )\n", szfilename );
	if (NULL == level_load( 0, szfilename ))
	{
		djMSG("GameLevelSetup(): error loading level %s.\n", szfilename );
		return;
	}
	g_pLevel = apLevels[0];

	pBackground = new djImage;
	if (0!=pBackground->Load(g_pCurMission->GetLevel(ilevel)->m_szBackground))
	{
		DEL(pBackground);
	}


	// Clear out inventory
	InvClear();
	InvDraw();

	g_ThingFactory.PerLevelInitialize();

	parse_level();


}
/*-----------------------------------------------------------*/

// Per-game cleanup
void GameCleanup()
{
	// Empty the inventory completely
	InvEmpty();
	// Delete game background image
	DEL(pBackground);
	// Delete all levels
	level_done();
}

// Per-level cleanup
void GameLevelCleanup()
{

}

// Final cleanup
void GameFinalCleanup()
{
	// Unload the game sounds (FIXME)
}
/*-----------------------------------------------------------*/
int game_startup()
{
	// FIXME: Where to determine this?
	TRACE( "game_startup()\n" );
	TRACE( "game_startup(): Playing game [%s]\n", g_pCurMission->GetName() );
	
	// Per game setup
	GameSetup();
	
	
	TRACE( "game_startup(): GameDrawSkin()\n" );
	GameDrawSkin();
	GraphFlip();

	// Per level setup (fixme, should this get called from withing per-game setup?
	GameLevelSetup();
	
	TRACE("game_startup(): GameDrawView()\n");
	// FIXME: Is this necessary?
	GameDrawView();
	// Draw inventory
	InvDraw();
	TRACE("game_startup(): update_score()\n");
	update_score(0);
	TRACE("game_startup(): draw_health()\n");
	update_health(0);
	
	g_bGameRunning = true;

	GameDrawFirepower();
	
	GraphFlip();
	
	// try maintain a specific frame rate
	float fTimeFrame = (1.0f / FRAME_RATE);
	
	float fTimeFirst = djTimeGetTime();
	
	// Start out by being at next time
	float fTimeNext = djTimeGetTime();
	float fTimeNow = fTimeNext;
	int iFrameCount=0;
	while (g_bGameRunning)
	{
		fTimeNow = djTimeGetTime();
		bool bForceUpdate = false;
		// If we're already behind, just force us to get there
		if (fTimeNext < fTimeNow)
		{
			//printf( "slow\n" );
			fTimeNext = fTimeNow;
			bForceUpdate = true;
		}

		int iEscape=0;
		while (fTimeNow<fTimeNext || bForceUpdate) // until we reach next frames time
		{
			// FIXME: THIS WILL HOG THE CPU
			
			// poll keys
			djiPoll();
			// only register key presses here, so that key presses aren't missed.
			if (g_iKeys[DJKEY_UP])    key_up = 1;
			if (g_iKeys[DJKEY_DOWN])  key_down = 1;
			if (g_iKeys[DJKEY_LEFT])  key_left = 1;
			if (g_iKeys[DJKEY_RIGHT]) key_right = 1;
			if (g_iKeys[DJKEY_CTRL])  key_ctrl = 1;
			if (g_iKeys[DJKEY_ALT])   key_shoot = 1;
			if (g_iKeys[DJKEY_P])     key_ctrl = 1;
			if (g_iKeys[DJKEY_ESC])   iEscape = 1;

			
			// Debug: hurt self
			static bool b = false;
			bool bOld = b;
			if (g_iKeys[DJKEY_H]) b = true; else b = false;
			if (b && !bOld)
				update_health(-1);

			// Debug: toggle display of debug info
			if (g_iKeys[DJKEY_D] && !g_iKeysLast[DJKEY_D]) bShowDebugInfo = !bShowDebugInfo;

			
			fTimeNow = djTimeGetTime();
			bForceUpdate = false;
		}
		// FIXME: time next should be calculated more absolutely, not relatively.
//		fTimeNext = fTimeNow + fTimeFrame;
		fTimeNext = fTimeNext + fTimeFrame;
		
		//-- ESC - Pop up the in-game menu
		if (iEscape==1)
		{
			int imenu = do_menu( &gameMenu );
			if (imenu == 4)
				g_bGameRunning = false;
			
			// Redraw everything that needs to be redrawn
			GameDrawSkin();
			GraphFlipView( VIEW_WIDTH );
			update_health( 0 );
			update_score( 0 );
			GameDrawFirepower();
			InvDraw();
			GraphFlip();
		}
		

		// Make a simple FPS display for debug purposes
		float fTimeRun;
		fTimeRun = fTimeNow - fTimeFirst;
		iFrameCount++;
		static char sbuf[1024];
		sprintf( sbuf, "%.2f", (float)iFrameCount / fTimeRun );
		if (iFrameCount==60)
		{
			iFrameCount /= 2;
			fTimeFirst += (fTimeRun/2);
		}
		djgDrawImage( pVisBack, pSkinGame, 0, 0, 0, 0, 196, 8 );
		GraphDrawString( pVisBack, g_pFont8x8, 0, 0, (unsigned char*)sbuf );

		// update
		float fT1 = djTimeGetTime();
		GameHeartBeat();
		float fT2 = djTimeGetTime();
		afTimeTaken.push_back((fT2 - fT1)*1000.0f);
		if (afTimeTaken.size()>MAX_DEBUGGRAPH)
			afTimeTaken.erase(afTimeTaken.begin());


		// FIXME: This behaviour is incorrect. A keyup inside a 2nd-time round
		// frame keypoll is not getting registered before the 2nd frame updates
		// and draws (huh?) (try move left/right only one block. It's difficult)
		key_up     = g_iKeys[DJKEY_UP];
		key_down   = g_iKeys[DJKEY_DOWN];
		key_left   = g_iKeys[DJKEY_LEFT];
		key_right  = g_iKeys[DJKEY_RIGHT];
		key_ctrl   = g_iKeys[DJKEY_CTRL];
		key_ctrl  |= g_iKeys[DJKEY_P];
		key_shoot  = g_iKeys[DJKEY_ALT];
		
		// ensure we don't leave the borders of the level
		// fixme; is this still necessary what with the (other functions)
		x = MAX( MIN(x,126), 1 );
		y = MAX( MIN(y, 99), 2 );

	} // while (game running)

	TRACE("game_startup(): main game loop exitted.\n");

	GameCleanup();
	return score;
}
/*-----------------------------------------------------------*/
void GameHeartBeat()
{
	CThing * pThing;
	int n, i, j;
	int ifoo;
	ifoo = key_up;

	// Check if we're about to teleport
	if (iTeleportCountdown != 0)
	{
		iTeleportCountdown--;
		if (iTeleportCountdown == 0)
		{
			relocate_hero( teleport_x, teleport_y );
		}
	}
	
	//update animation counts:
	anim4_count++;
	if (anim4_count>3)
		anim4_count = 0;
	
	foocount++;
	if (foocount>1)
		foocount = 0;
	
	//not jumping but about to be, then dont left/right move
	if (!((key_ctrl) && (hero_mode != MODE_JUMPING))) {
		if (key_left)
		{
			key_left = 0;
			move_hero(-1,0);
		}
		if (key_right)
		{
			key_right = 0;
			move_hero(1,0);
		}
	}
	
	//mode-specific handling
	switch (hero_mode)
	{
	case MODE_NORMAL:
		//fall:
		n = move_hero(0,1);
		
		// standing still and pressing 'up': (just pressing up?)
		if (key_up)
		{
			key_up = 0; // huh?
			
			// Level cheat key (Ctrl+L) (Fixme, remove in final version)
			if (g_iKeys[DJKEY_L])
			{
				SetLevel(GetCurrentLevel()+1);
				return;
			}
			// Give-all-keys cheat key
			if (g_iKeys[DJKEY_K])
			{
				vector<int> anKeysHave;
				for ( i=0; i<InvGetSize(); i++ )
				{
					if (InvGetItem(i)->GetTypeID()==TYPE_KEY)
					{
						CKey *pKey = (CKey*)InvGetItem(i);
						anKeysHave.push_back(pKey->GetID());
					}
				}
				for ( i=1; i<=4; i++)
				{
					unsigned int j;
					bool bHave = false;
					for ( j=0; j<anKeysHave.size(); j++ )
					{
						if (i==anKeysHave[j])
							bHave = true;
					}
					if (!bHave)
					{
						CKey *pKey = new CKey;
						pKey->SetType(TYPE_KEY);
						pKey->SetSprite(0, 116+i);
						pKey->Initialize(0, 116+i);
						InvAdd(pKey);
					}
				}
			}
			
			// Check if you're on anything funny, like an exit
			for ( i=0; i<(int)g_apThings.size(); i++ )
			{
				CThing *pThing = g_apThings[i];
				if (pThing->InBounds(x*16+x_small*8, y*16-16))
				{
					int iRet = pThing->Action();
					switch (iRet)
					{
					case ACTION_EXIT:
						// we do nothing here, but we must return
						return;
					case ACTION_TELEPORT:
						if (iTeleportCountdown == 0) // If not already busy teleporting
						{
							// Locate the other teleporter
							bool bFound = false;
							unsigned int j;
							for ( j=0; j<g_apThings.size(); j++ )
							{
								CThing *pThing2 = g_apThings[j];
								if (pThing2 != pThing) // If this isn't the teleporter we're standing on
								{
									if (pThing2->GetTypeID() == TYPE_TELEPORTER) // If this is a teleporter
									{
										// If second teleporter ID is same as this one
										if (pThing2->ID() == pThing->ID())
										{
											teleport_x = pThing2->m_x;
											teleport_y = pThing2->m_y;
											bFound = true;
											iTeleportCountdown = 12; // start the teleporter countdown
										}
									}
								}
							} // j
							if (!bFound)
							{
								djMSG( "Warning: No matching teleport found for %d\n", pThing->ID() );
							}
						}
						break;
					}
				}
			} // i
			
			
		}
		
		
		if (key_ctrl)
		{
			key_ctrl = 0;
			if (n) { // if hero wasnt busy falling
				hero_mode = MODE_JUMPING;
				jump_pos = 0;
				hero_picoffs=1;
			}
		}
		break;
		
	case MODE_JUMPING:
		n = move_hero(0,pJumpInfo->jump_diffs[jump_pos]);
		if (n == 1) { //cancel jump, fell on block (x 1 ret?)
			hero_mode = MODE_NORMAL;
			hero_picoffs = 1;
		}
		jump_pos++;
		if (jump_pos >= pJumpInfo->size)
			hero_mode = MODE_NORMAL;
		if (key_ctrl)
		{
			key_ctrl = 0;
		}
		break;
		
	}

	// Check for bullets that have gone out of the view
	for ( i=0; i<(int)g_apBullets.size(); i++ )
	{
		CBullet *pBullet = g_apBullets[i];
		int x1 = pBullet->dx<0 ? pBullet->x + 8 : pBullet->x;
		if (!OVERLAPS_VIEW(
			x1,
			pBullet->y,
			x1 + 7,
			pBullet->y+BULLET_HEIGHT-1))
		{
			DEL(pBullet);
			g_apBullets.erase(g_apBullets.begin() + i);
			i--;
		}
	}
	
	// Check for bullet collisions with things (e.g. shootable boxes)
	for ( i=0; i<(int)g_apBullets.size(); i++ )
	{
		CBullet *pBullet = g_apBullets[i];
		for ( j=0; j<(int)g_apThings.size(); j++ )
		{
			CThing *pThing = g_apThings[j];
			if (pThing->IsShootable())
			{
				int x1 = pBullet->dx<0 ? pBullet->x + 8 : pBullet->x;
				if (pThing->OverlapsShootArea(
					x1,
					pBullet->y,
					x1 + 7,
					pBullet->y+BULLET_HEIGHT-1))
					/*
				if (pThing->OverlapsBounds(
					pBullet->x,
					pBullet->y,
					pBullet->x+BULLET_WIDTH-1,
					pBullet->y+BULLET_HEIGHT-1))
					*/
				{
					int nRet = pThing->OnHeroShot();
					if (nRet==THING_DIE)
					{
						delete pThing;
						g_apThings.erase(g_apThings.begin() + j);
						j--;
					}
					else if (nRet==THING_REMOVE)
					{
						g_apThings.erase(g_apThings.begin() + j);
						j--;
					}
					
					// delete bullet i
					g_apBullets.erase(g_apBullets.begin() + i);
					i--;
					goto NextBullet1;
				}
			}
		} // j
		
NextBullet1:
		;

	}

	// Create new bullets
	static int bulletcount = 0;
	if (key_shoot)
	{
		if (bulletcount==0 && (int)g_apBullets.size()<g_nFirepower)
		{
			bulletcount = 4; // can't shoot for 4 frames
			CBullet *pBullet = new CBullet;
			pBullet->x = x * 16 + (hero_dir==1 ? 16 : -16) + x_small*8;
			pBullet->y = y * 16 - 2;
			pBullet->dx = (hero_dir==0 ? -16 : 16);
			g_apBullets.push_back(pBullet);
			djSoundPlay( g_iSounds[SOUND_SHOOT] );
		}
	}
	if (bulletcount != 0)
		bulletcount--;


	for ( i=0; i<(int)g_apBullets.size(); i++ )
	{
		CBullet *pBullet = g_apBullets[i];
		int nXOld = pBullet->x;
		int nYOld = pBullet->y;
		// Update bullet
		pBullet->Tick();

		// check if bullet touching anything solid
		int x1 = pBullet->dx<0 ? pBullet->x + 8 : pBullet->x;
		if (CheckCollision(
			x1,
			pBullet->y,
			x1 + 7,
			pBullet->y+BULLET_HEIGHT-1, true))
		{
			AddThing(CreateExplosion((nXOld + (pBullet->dx<0 ? 0 : 0)), nYOld-4));
			g_apBullets.erase(g_apBullets.begin() + i);
			i--;
			goto NextBullet3;
		}


NextBullet3:
		;

	
	}


	// Interact with "things"
	// Check if you're on anything funny, like an exit
	for ( i=0; i<(int)g_apThings.size(); i++ )
	{
		CThing *pThing = g_apThings[i];
		if (pThing->OverlapsBounds(x*16+x_small*8, y*16-16))
		{
			int nRet = pThing->HeroOverlaps();
			if (nRet==THING_DIE)
			{
				delete pThing;
				g_apThings.erase(g_apThings.begin() + i);
				i--;
				pThing = NULL;
			}
			else if (nRet==THING_REMOVE)
			{
				g_apThings.erase(g_apThings.begin() + i);
				i--;
				pThing = NULL;
			}
		}
		// If thing wasn't deleted on HeroOverlaps, check bounds enter/leave
		if (pThing!=NULL)
		{
			// Test if entering or leaving action bounds box
			if (pThing->InBounds(x*16+x_small*8, y*16-16))
			{
				if (!pThing->IsHeroInside())
					pThing->HeroEnter();
			}
			else // not in bounds
			{
				if (pThing->IsHeroInside())
					pThing->HeroLeave();
			}
		}
	}


	
	// Drop all objects that can fall
	for ( i=0; i<(int)g_apThings.size(); i++ )
	{
		pThing = g_apThings[i];
		// if (object falls) && (nothing below it) && (inview)
		if (pThing->Falls())
		{
			// If thing is in visible view
			// FIXME: Small border around view?
			if (pThing->IsInView())
			{
				// if (nothing below it)
				if ( !check_solid( pThing->m_x, pThing->m_y + 1 ) )
				{
					pThing->m_y += 1;
					// if object falls off bottom of level
					if (pThing->m_y >= 100)
					{
						// delete this object!!!
						delete pThing;
						g_apThings.erase(g_apThings.begin() + i);
						i--;
					}
				}
			}
		}
	}


	// "Tick" (update) all objects
	for ( i=0; i<(int)g_apThings.size(); i++ )
	{
		pThing = g_apThings[i];
		// FIXME: THING_REMOVE?
		if (pThing->Tick()==THING_DIE)
		{
			// Delete this
			delete pThing;
			g_apThings.erase(g_apThings.begin() + i);
			i--;
		}
	}






	// Redraw the screen according to the current game state
	GameDrawView();

	
	if ( nHurtCounter > 0 )
		nHurtCounter--;

	// Flip the back buffer onto the front
	GraphFlip();
}

void update_health(int health_diff)
{
	int i;
	health += health_diff;
	health = MIN( MAX(health,0), 10 );
	if (health<=0)
		Die();

	// If busy jumping up and something hurts hero, stop the jump
	if (health_diff<0 && hero_mode==MODE_JUMPING)
	{
		jump_pos = 0;
		hero_mode = MODE_NORMAL;
	}

	char szHealth[16];
	for ( i=0; i<10; i++ )
	{
		// 170 = health; 169 = not health
		szHealth[9-i] = (i<health?170:169);
	}
	szHealth[10] = 0;
	GraphDrawString( pVisBack, g_pFont8x8, HEALTH_X, HEALTH_Y, (unsigned char*)szHealth );
}

void Die()
{

	// Reset to beginning of current level
	SetLevel(ilevel);
//	g_iLives--;
//	if (g_iLives<=0)
//	{
//		g_bGameRunning = false;
//		return;
//	}
		// FIXME: SHOULD RESTART LEVEL OR SOMETHING
	// Set "nHurtCounter" so we have a bit of time to get back on our feet
	//HeroSetHurting(true);
	GameDrawFirepower();
}

void HeroSetHurting(bool bReset)
{
	if (bReset || nHurtCounter==0)
		nHurtCounter = 16;
}

bool HeroIsHurting()
{
	return nHurtCounter!=0;
}

void update_score(int score_diff, int nFloatingScoreX, int nFloatingScoreY)
{
	char score_buf[16];
	score += score_diff;
	sprintf( score_buf, "%10d", (int)score );
	
	if (pSkinGame)
	{
		// Clear behind the score with part of the game skin
		djgDrawImage( pVisBack, pSkinGame, SCORE_X, SCORE_Y, SCORE_X, SCORE_Y, 10*8, 8 );
	}
	else
	{
		// Oops, no game skin, just clear behind the score with plain black
		djgSetColorFore( pVisBack, djColor(0,0,0) );
		djgDrawBox( pVisBack, SCORE_X, SCORE_Y, 10*8, 8 );
	}
	// Display score
	GraphDrawString( pVisBack, g_pFont8x8, SCORE_X, SCORE_Y, (unsigned char*)score_buf );
	
	// If requested, create a "floating score" display
	if (nFloatingScoreX!=-1 && nFloatingScoreY!=-1)
		AddThing(CreateFloatingScore(nFloatingScoreX, nFloatingScoreY, score_diff));
}
/*-----------------------------------------------------------*/
//returns 0 if could move
//returns 1 if couldnt move (y)
//returns 2 if coundnt move (x)
int move_hero(int xdiff, int ydiff, bool bChangeLookDirection)
{
	int n,ret;
	bool bsolid;
	ret = 0;
	
	// don't do any moving if you're about to teleport
	if (iTeleportCountdown != 0)
		return 1;
	
	if (xdiff)
	{
		//simple direction reverse (unless bChangeLookDirection is false, which means we're probably on a conveyor or something)
		if (bChangeLookDirection && (((hero_dir == 1) & (xdiff == -1)) | ((hero_dir == 0) & (xdiff == 1))))
		{
			hero_picoffs = 0;
			hero_dir = (xdiff+1)/2;  //( -1 --> 0 : 1 --> 1 )
			return 0;
		}
		//facing right, must also have pressed right
		//if (hero_dir)
		if (xdiff>0)
		{
			bsolid = false;
			if ( x_small == 0 ) {
				bsolid = check_solid( x + 1, y ) | check_solid( x + 1, y - 1 );
			}
			ret = 2;
			if (  (!(bsolid)) && ( (x_small) | (!(bsolid)) )  ) {
				ret = 0;
				x += (xdiff * x_small);
				x_small ^= 1;
				
				if (((x-xo)==(VIEW_WIDTH - 5)) & (x_small)) {
					xo_small = 1;
				}
				if (((x-xo)==(VIEW_WIDTH - 4) )) {
					xo++;
					xo_small = 0;
				}
				if ( (xo + xo_small) > 128 - VIEW_WIDTH )
				{
					xo = 128 - VIEW_WIDTH;
					xo_small = 0;
				}
				
			}
		}
		else
		{ //facing left, must have pressed left
			bsolid = false;
			if (!(x_small))
			{
				bsolid = check_solid( x - 1, y ) | check_solid( x - 1, y - 1 );
			}
			ret = 2;
			if ((!(bsolid)) & ((x_small) | (!(bsolid))) ) {
				ret = 0;
				x_small ^= 1;
				x += (xdiff * x_small);
				if (((x-xo)==4) & (!(x_small))) {
					xo_small = 0;
				}
				if (((x-xo)==3) & (x_small)) {
					xo--;
					xo_small = 1;
					if (xo < 0)
					{
						xo = 0;
						xo_small = 0;
					}
				}
			}
			
		}
	}
	
	if (ydiff)
	{
		if (ydiff == 1)
			n = 1;      // falling, check below us
		else
			n = -2;     // going up, check above hero's head
		
		// also stop hero falling if at bottom of screen
		bsolid = check_solid( x, y + n ) | check_solid( x + x_small, y + n );
		
		ret = 1;
		if (!bsolid) {
			ret = 0;
			y += ydiff;
			if (y-yo<4) yo--;
			if (y-yo>7) yo++;
			if ( yo < 0 )
				yo = 0;
			if ( yo > 100 - 10 )
				yo = 100 - 10;
		}
	}
	if ((hero_mode != MODE_JUMPING) && (ret == 0)) {
		if (foocount == 0)
			hero_picoffs++;
		if (hero_picoffs>3) hero_picoffs = 0;
	}
	
	return(ret);
}

void DrawThingsAtLayer(EdjLayer eLayer)
{
	CThing *pThing;
	unsigned int i;
	for ( i=0; i<g_apThings.size(); i++ )
	{
		pThing = g_apThings[i];
		if (pThing->Layer()==eLayer && pThing->IsInView())
			pThing->Draw();
	}
}

void GameDrawView()
{
	int i,j,a,b,xoff,yoff;
	int anim_offset = 0;
	unsigned char *tempptr;
	CThing * pThing;
	
	// Clear view background
	if (pBackground)
		djgDrawImage(pVisView, pBackground, 0, 0, 16, 16, VIEW_WIDTH*16, VIEW_HEIGHT*16);
	else
		djgClear(pVisView);
	
	//(10 seconds got to just after coke can, purple lab)
	tempptr = (unsigned char *)(g_pLevel) + yo*512+(xo<<2);
	yoff = 200+16;
	//  c=2;
	for ( i=0; i<10; i++ )
	{
		//  d=24;
		xoff = -xo_small+2;
		for ( j=0; j<VIEW_WIDTH+xo_small; j++ )
		{
			// BLOCK[2,3] -> background block
			a = *(tempptr+2);
			b = *(tempptr+3);
			// Animated block?
			anim_offset = (GET_EXTRA( a, b, 4 ) & FLAG_ANIMATED) ? anim4_count : 0;

			// draw background block
			//djgDrawImage( pVisView, g_pCurMission->GetSpriteData(a)->m_pImage, ((b+anim_offset)%16)*16, ((b+anim_offset)/16)*16, xoff*8,16+i*16,16,16 );
			if ((a | b) != 0)
				DRAW_SPRITE16A(pVisView, a, b+anim_offset, xoff*8, 16+i*16);
			
			// BLOCK[0,1] -> foreground block
			a = *(tempptr);
			b = *(tempptr+1);
			// Animated block?
			anim_offset = (GET_EXTRA( a, b, 4 ) & FLAG_ANIMATED) ? anim4_count : 0;

			// draw foreground block, unless its (0,0)
			if ((a | b) != 0)
				DRAW_SPRITE16A(pVisView, a, b+anim_offset, xoff*8, 16+i*16);
			
			xoff+=2;
			tempptr += 4;
		}
		yoff += 16;
		tempptr += (512 - ((VIEW_WIDTH+xo_small)<<2));
	}
	
	// Draw bottom most layers, then falling objects, then higher layers
	// FIXME: Make this uniform, damnit
	DrawThingsAtLayer(LAYER_BOTTOM);
	
	// Draw up to middle
	DrawThingsAtLayer(LAYER_2);
	DrawThingsAtLayer(LAYER_MIDDLE);
	
	// draw hero, but flash if he is currently hurt
	if ((nHurtCounter == 0) || (nHurtCounter%3 != 0))
	{
		xoff = (x_small - xo_small)+1 + ((x-xo)<<1);
		yoff = 200+16+(y-yo-1)*16;
		
		xoff = ((x_small - xo_small)+1)*8 + (x-xo) * 16;
		yoff = 16 + (y-yo-1) * 16;
		
		xoff = (x_small - xo_small)+1 + ((x-xo)<<1);
		xoff *= 8;
		DRAW_SPRITE16A(pVisView,4,  hero_dir*16+hero_picoffs*4,xoff   ,yoff   );
		DRAW_SPRITE16A(pVisView,4,2+hero_dir*16+hero_picoffs*4,xoff   ,yoff+16);
		DRAW_SPRITE16A(pVisView,4,1+hero_dir*16+hero_picoffs*4,xoff+16,yoff   );
		DRAW_SPRITE16A(pVisView,4,3+hero_dir*16+hero_picoffs*4,xoff+16,yoff+16);
	}
	
	// Draw layers past middle
	DrawThingsAtLayer(LAYER_4);
	DrawThingsAtLayer(LAYER_TOP);
	
	// Draw bullets
	DrawBullets();
	
	if (bShowDebugInfo)
	{
		djgSetColorFore(pVisView,djColor(255,50,255));
		for ( i=0; i<(int)g_apBullets.size(); i++ )
		{
			CBullet *pBullet = g_apBullets[i];
			if (OVERLAPS_VIEW(pBullet->x, pBullet->y, pBullet->x+8, pBullet->y+8))
			{
				djgDrawRectangle(pVisView,
					WORLDX2VIEW(pBullet->x),
					WORLDY2VIEW(pBullet->y),
					BULLET_WIDTH,
					BULLET_HEIGHT);
			}
		}

		// Time taken graph
		float fScale = 2.0f;
		djgSetColorFore(pVisView,djColor(150,150,150));
		for ( float fAxis=0; fAxis<=50.0f; fAxis += 10.0f )
		{
			djgDrawRectangle(pVisView,
				10, 170 - (int)(fAxis * fScale),
				MAX_DEBUGGRAPH, 1);
		}
		djgSetColorFore(pVisView,djColor(0,255,0));
		for ( i=0; i<(int)afTimeTaken.size(); i++ )
		{
			float f = afTimeTaken[i];
			djgDrawRectangle(pVisView,
				10 + i,170 - (int)(f * fScale),
				1, (int)(f * fScale));
		}

		
		int nNumVisible = 0;
		// DEBUG: Draw "action" bounding boxes
		for ( i=0; i<(int)g_apThings.size(); i++ )
		{
			pThing = g_apThings[i];

			// Draw "visible" bounds (red)
			// "Visible" box (red)
			djgSetColorFore(pVisView,djColor(255,0,0));
			if (pThing->IsInView())
			{
				nNumVisible++;
				djgDrawRectangle(pVisView,
					CALC_XOFFSET(pThing->m_x,pThing->m_xsmall)+pThing->m_iVisibleX1,
					CALC_YOFFSET(pThing->m_y)+pThing->m_iVisibleY1,
					(pThing->m_iVisibleX2-pThing->m_iVisibleX1)+1,
					(pThing->m_iVisibleY2-pThing->m_iVisibleY1)+1);
				djgDrawRectangle(pVisView,
					1 + (CALC_XOFFSET(pThing->m_x,pThing->m_xsmall)+pThing->m_iVisibleX1),
					1 + (CALC_YOFFSET(pThing->m_y)+pThing->m_iVisibleY1),
					(pThing->m_iVisibleX2-pThing->m_iVisibleX1)-1,
					(pThing->m_iVisibleY2-pThing->m_iVisibleY1)-1);
			}

			// Draw action bounds (cyan=overlapping, white=inside, yellow=not interacting)
			if (pThing->OverlapsBounds(x*16+x_small*8, y*16-16))
			{
				if (pThing->IsHeroInside())
					pThing->DrawActionBounds(djColor(255,255,255));
				else
					pThing->DrawActionBounds(djColor(0,255,255));
			}
			else
				pThing->DrawActionBounds(djColor(255,255,0));
			// Draw solid bounds (green)
			if (pThing->m_bSolid)
			{
				djgSetColorFore( pVisView, djColor(0,255,0) );
				djgDrawRectangle( pVisView,
					CALC_XOFFSET(pThing->m_x,pThing->m_xsmall) + pThing->m_iSolidX1,
					CALC_YOFFSET(pThing->m_y) + pThing->m_iSolidY1,
					(pThing->m_iSolidX2-pThing->m_iSolidX1)+1,
					(pThing->m_iSolidY2-pThing->m_iSolidY1)+1 );
			}
		}

		GraphDrawString(pVisView, g_pFont8x8, 32, 16, (unsigned char*)"Debug info on (D)" );
		char buf[64];
		sprintf(buf, "%d things", g_apThings.size());
		GraphDrawString(pVisView, g_pFont8x8, 32, 24, (unsigned char*)buf );
		sprintf(buf, "%d visible", nNumVisible);
		GraphDrawString(pVisView, g_pFont8x8, 32, 32, (unsigned char*)buf );
		sprintf(buf, "[%d,%d] [%d firepower]", x, y, g_nFirepower);
		GraphDrawString(pVisView, g_pFont8x8, 32, 40, (unsigned char*)buf );

	}
	
	// Flip the off-screen world viewport onto the backbuffer
	GraphFlipView( VIEW_WIDTH );
}

void parse_level(void)
{
	int i, j;
	// parse the level (for doors, keys, hero starting position etc.)
	for ( i=0; i<100; i++ )
	{
		for ( j=0; j<128; j++ )
		{
			sprite_factory( 0, 0, j, i, 0, true );
			sprite_factory( 0, 0, j, i, 1, true );
		}
	}
}

void sprite_factory( unsigned char a, unsigned char b, int ix, int iy, int ifore, bool bfromlevel )
{
	unsigned char * plevel;
	unsigned char   b0, b1;
	bool            bWipeSprite = false;
	
	plevel = g_pLevel + 4 * (iy * 128 + ix) + (ifore * 2);
	
	if ( bfromlevel )
	{
		b0 = *(plevel    );
		b1 = *(plevel + 1);
	}
	else
	{
		b0 = a;
		b1 = b;
	}
	
	// Check if this block is a valid sprite
	if (g_pCurMission->GetSpriteData( b0 ) == NULL)
	{
		djMSG( "WARNING: INVALID SPRITESET ID FOUND (%d,%d) AT (%d,%d,%d)\n",
			b0, b1, ix, iy, ifore );
		if ( bfromlevel )
		{
			*(plevel    ) = 0;
			*(plevel + 1) = 0;
		}
		return;
	}
	
	// First attempt to allocate from "CThing" factory
	int nTypeID = g_pCurMission->GetSpriteData(b0)->m_type[b1];
	CThing *pThing = g_ThingFactory.Allocate(nTypeID);
	if (pThing)
	{
		pThing->SetLocation(ix, iy, 0, 0, 1, 1);
		pThing->SetSprite(b0, b1);
		pThing->Initialize(b0, b1);
		AddThing(pThing);
		bWipeSprite = true;
	}
	else // Special types
	{
		switch (nTypeID)
		{
		case TYPE_HEROSTART:
			// hero starting position
			relocate_hero( ix, iy );
			
			hero_dir = 1;
			bWipeSprite = true;
			break;
		} //switch (block type)
	}

	// Remove from physical level data, as this has been converted
	// to a thing, or some other action taken
	if ( (bWipeSprite == true) && (bfromlevel == true) )
	{
		*(plevel    ) = 0;
		*(plevel + 1) = 0;
	}
	
}

void GameDrawSkin()
{
	// Draw the game skin
	if (pSkinGame)
		djgDrawImage( pVisBack, pSkinGame, 0, 0, 320, 200 );
}

int GetCurrentLevel()
{
	return ilevel;
}

void SetLevel(int nLevel)
{
	ilevel = nLevel;
	djiWaitForKeyUp( DJKEY_L );
	if ( ilevel >= g_pCurMission->NumLevels() )
	{
		// You have finished the game!!! Woohoo!!!!
		djMSG( "-----------------------------------------\n" );
		djMSG( "You have finished the game!!! Woohoo!!!!.\n" );
		djMSG( "-----------------------------------------\n" );
		g_bGameRunning = false;
		return;
	}
	GameLevelSetup();
	
	return;
}

void DrawBullets()
{
	unsigned int i;
	for ( i=0; i<g_apBullets.size(); i++ )
	{
		CBullet *pBullet = g_apBullets[i];
		if (OVERLAPS_VIEW(pBullet->x, pBullet->y, pBullet->x+BULLET_WIDTH, pBullet->y+BULLET_HEIGHT))
		{
			pBullet->Draw();
		}
	}
}

void relocate_hero( int xnew, int ynew )
{
	x = xnew;
	y = ynew;
	x_small = 0;
	xo = MAX( x - int(VIEW_WIDTH / 2), 0 );
	yo = MAX( y - 6, 0 );
	xo = MIN( xo, 128 - VIEW_WIDTH );
	yo = MIN( yo, 100 - VIEW_HEIGHT );
}

bool check_solid( int ix, int iy, bool bCheckThings )
{
	int i;
	
	// Create an invisible "border" around the level. Handy catch-all for things going out of bounds.
	if ( ix<1 || iy<1 || ix>126 || iy>98 ) return true;
	
	// FIXME: This is not speed-optimal (does it really matter?)

	// If foreground block is solid
	if (CHECK_SOLID( LEVCHAR_FOREA(ix,iy), LEVCHAR_FOREB(ix,iy) )) return true;
	// If background block is solid	
	if (CHECK_SOLID( LEVCHAR_BACKA(ix,iy), LEVCHAR_BACKB(ix,iy) )) return true;

	// Check "things"
	if (bCheckThings)
	{
		CThing *pThing = NULL;
		for ( i=0; i<(int)g_apThings.size(); i++ )
		{
			pThing = g_apThings[i];
			if (pThing->m_bSolid)
			{
				if (OVERLAPS(
					ix*16, iy*16, ix*16+15, iy*16+15,
					pThing->m_x*16 + pThing->m_iSolidX1,
					pThing->m_y*16 + pThing->m_iSolidY1,
					pThing->m_x*16 + pThing->m_iSolidX2,
					pThing->m_y*16 + pThing->m_iSolidY2))
				{
					return true;
				}
			}
			//if (pThing->CheckSolid(ix, iy))
			//	return true;
		}
	}
	
	return false;
}

bool CheckCollision(int x1, int y1, int x2, int y2, bool bBullet)
{
	int i, j;
	CThing *pThing;
	int nX1 = x1 / 16;
	int nY1 = y1 / 16;
	int nX2 = x2 / 16;
	int nY2 = y2 / 16;
	for ( i=nX1; i<=nX2; i++ )
	{
		for ( j=nY1; j<=nY2; j++ )
		{
			if (check_solid(i, j, false))
			{
				if (OVERLAPS(x1, y1, x2, y2, i*16, j*16, i*16+15, j*16+15))
					return true;
			}
		}
	}
	for ( i=0; i<(int)g_apThings.size(); i++ )
	{
		pThing = g_apThings[i];
		if (pThing->m_bSolid)
		{
			if (OVERLAPS(
				x1, y1, x2, y2,
				pThing->m_x*16 + pThing->m_iShootX1,
				pThing->m_y*16 + pThing->m_iShootY1,
				pThing->m_x*16 + pThing->m_iShootX2,
				pThing->m_y*16 + pThing->m_iShootY2))
			{
				if (bBullet && pThing->IsShootable())
				{
					int nRet = pThing->OnHeroShot();
					if (nRet==THING_DIE)
					{
						delete pThing;
						g_apThings.erase(g_apThings.begin() + i);
					}
					else if (nRet==THING_REMOVE)
					{
						g_apThings.erase(g_apThings.begin() + i);
					}
				}
				return true;
			}
		}
	}

	return false;
}


int GameLoadSprites()
{
	int i;
	
	// Load the raw sprite data
	i = g_pCurMission->LoadSprites();
	if ( i != 0 )
	{
		djMSG( "GameLoadSprites(): Unable to load raw sprite data\n" );
		return i;
	}
	
	// Convert sprite data to format useful for ggi
	// FIXME: This doesn't happen at the moment. It currently doesn't
	// need to do anything of the sort, but should in future to take
	// advantage of (for example) hardware acceleration, but at the
	// moment it just draws the sprites straight out of the djImage.

	// i iterates through the 256 possible ID's for spritesets
	for ( i=0; i<256; i++ )
	{
		CSpriteData * pSpriteData;
		
		pSpriteData = g_pCurMission->GetSpriteData( i );
		if (pSpriteData != NULL)
		{
			// The sprite data must now be converted to a format useful for ggi
		} // if
	} // i
	
	return 0;
}

void GameDrawLives()
{
	int i;
	for ( i=0; i<LIVES_MAX; i++ )
	{
		// draw life
		if (i<g_iLives)
			DRAW_SPRITE16A( pVisBack, 4, 127, LIVES_X-i*16, LIVES_Y );
//		else // draw background
//			djgDrawImage( pVisBack, pSkinGame, LIVES_X-i*16, LIVES_Y, LIVES_X-i*16, LIVES_Y, 16, 16 );
	}
}

void GameDrawFirepower()
{
	int i;
	GameDrawLives(); // <-- fixme
	djgDrawImage( pVisBack, pSkinGame, LIVES_X-4*16, LIVES_Y, LIVES_X-4*16, LIVES_Y, 16*5, 16 );
	for ( i=0; i<g_nFirepower; i++ )
	{
		DRAW_SPRITE16A( pVisBack, 5, 0, LIVES_X+i*16-4*16, LIVES_Y );
	}
}

void AddThing(CThing *pThing)
{
	if (pThing!=NULL) // This can be NULL in some cases
	{
		g_apThings.push_back(pThing);
		pThing->OnAdded();
	}
}


void HeroModifyFirepower(int nDiff)
{
	g_nFirepower += nDiff;
	GameDrawFirepower();
}
