/*!
\file    game.h
\brief   The control of the game itself
\author  David Joffe
*/
/*--------------------------------------------------------------------------*/
/* David Joffe '95/07/28 */
/* Game playing stuff for djg */
/*--------------------------------------------------------------------------*/
#ifndef _GAME_H_
#define _GAME_H_

#include <vector>
using namespace std;

//! Visible blocks on X axis in game viewport
#define VIEW_WIDTH (12)
//! Visible blocks on Y axis in game viewport
#define VIEW_HEIGHT (10)

//! Tests if rectangle overlaps portion of world visible on screen (pixel coordinates)
#define OVERLAPS_VIEW(x1,y1,x2,y2) (OVERLAPS(x1,y1,x2,y2,16*xo,16*yo,16*(xo+VIEW_WIDTH)+(8*xo_small),16*(yo+VIEW_HEIGHT)))
//! Tests if point is directly inside portion of world visible on screen (pixel coordinates)
#define IN_VIEW(x,y) (INBOUNDS(x,y,xo*16,yo*16,16*(xo+VIEW_WIDTH)+(8*xo_small),16*(yo+VIEW_HEIGHT)))


class CThing;
extern vector<CThing *> g_apThings;

//! Add a \ref CThing to the current array of things in the level.
extern void AddThing(CThing *pThing);


#include "djsound.h"
enum EdjGameSounds
{
	SOUND_PICKUP=0,
	SOUND_SHOOT,
	SOUND_EXIT,
	SOUND_OPENDOOR,
	SOUND_MAX
};
extern SOUND_HANDLE g_iSounds[SOUND_MAX];

// x,y = hero's absolute positioning in level
// Fixme, these names have GOT to change!
extern int x;
extern int y;
extern int xo;
extern int yo;// xo,yo = top-left corner of view for scrolling
extern int x_small;
extern int xo_small; // x_small == 0 ? hero at x : hero at x + 8 pixels
//! Global 4-animation-frame animation count. This name must also change.
extern int anim4_count;

//! Jump modes
enum EJump
{
	JUMP_NORMAL,
	JUMP_POWERBOOTS
};
//! Set hero's current jump mode
extern void HeroSetJumpMode(EJump eJump);



//----- Game control

//! Set current level (0-based index)
extern void SetLevel(int nLevel);
//! Get current level (0-based index)
extern int GetCurrentLevel();
//! Lose a life
extern void Die();
//! The main game "tick" routine
extern void GameHeartBeat();

//----- Game screen redrawing

//! Draw the main game skin
extern void GameDrawSkin();
//! Draw hero's remaining "lives" (OBSOLETE?)
extern void GameDrawLives();
//! Draw the main game view
extern void GameDrawView();
//! Draw hero firepower
extern void GameDrawFirepower();

//! Once off initialization stuff
extern void GameInitialSetup();
//! Final cleanup
extern void GameFinalCleanup();

//! Per-game initialization
extern void GameSetup();
extern int GameLoadSprites(); // fixme , move into above func?
//! Per-level initialization
extern void GameLevelSetup();
//! Per-game cleanup
extern void GameCleanup();
//! Per-level cleanup
extern void GameLevelCleanup();

extern void relocate_hero( int xnew, int ynew );

//! Add some value to hero firepower
extern void HeroModifyFirepower(int nDiff);



//! Return true if solid. bCheckThings is an ugly hack which (fixme) should be fixed
extern bool check_solid( int ix, int iy, bool bCheckThings=true );

//! Return true if given rectangle [pixel coordinates] is colliding with anything solid
extern bool CheckCollision(int x1, int y1, int x2, int y2, bool bBullet=false);


extern int game_startup();
extern void parse_level();
extern void sprite_factory( unsigned char a, unsigned char b, int ix, int iy, int ifore, bool bfromlevel );

extern int  move_hero(int xdiff, int ydiff, bool bChangeLookDirection=true);

extern void update_score(int score_diff, int nFloatingScoreX=-1, int nFloatingScoreY=-1);
extern void update_health(int health_diff);

//! Set the hero state to "hurting", unless already hurting, unless bReset is set to true
extern void HeroSetHurting(bool bReset=false);
//! Return true if hero is "hurting" (flashing)
extern bool HeroIsHurting();

#endif
