/*--------------------------------------------------------------------------*/
/* David Joffe '95/07/20 */
// Pre-1995 : something similar from QBasic or Pascal, I can't remember ...
// 1995 : original DOS/EGA version
// 1998/12/23 : begin attempted Linux port
// 1999/05 : begin attempted Win32 port, ack
// 1999/12 : re-begin attempted Win32 port
// 2001/05 : begin SDL port; doxygen comments
/*--------------------------------------------------------------------------*/

#include <stdio.h>

#include <time.h>   // for srand()
#include <stdlib.h> // for rand()

#include <string.h>

#include "graph.h"

//- djg lib -------------
#include "djgamelib.h"
//-----------------------

#include "hiscores.h"
#include "menu.h"
#include "game.h"
#include "level.h"

#include "mission.h"

#include "credits.h"
#include "instructions.h"

/*--------------------------------------------------------------------------*/
djImage *pImgMain=NULL;

int  DaveStartup(bool bFullScreen);	// Application initialization
void DaveCleanup();					// Application cleanup
void StartNewGame();				// Start new game
void SelectMission();				// Select a mission

void DoMainMenu();					// Start main menu

/*--------------------------------------------------------------------------*/
// Main menu
struct SMenuItem mainMenuItems[] =
{
   { false, "{~~~~~~~~~~~~~~~~~}" },
   { true,  "|  Start gnu game |" },
   { true,  "|  Select Mission |" },
   { true,  "|  Restore game   |" },
   { true,  "|  Ordering info  |" },
   { true,  "|   (not!)        |" },
   { true,  "|  Instructions   |" },
   { true,  "|  Dont quit      |" },
   { true,  "|  High scores    |" },
   { true,  "|  Credits        |" },
   { true,  "|  Quit           |" },
   { false, "[~~~~~~~~~~~~~~~~~]" },
   { false, NULL }     
};
unsigned char mainMenuCursor[] = { 128, 129, 130, 131, 0 };
unsigned char mainMenuCursorSkull[] = { 161, 162, 163, 164, 0 };
CMenu mainMenu;
/*--------------------------------------------------------------------------*/
// This is the 'main' function. The big cheese.
int main( int argc, char** argv )
{
	// Check commandline args
	bool bfullscreen = false;
	if (argc > 1)
	{
		for ( int i=1; i<argc; i++ )
		{
			if (0 == strncmp( argv[i], "-f", 2 )) bfullscreen = true;
		}
	}
	else
	{
		// Show usage information
		printf( "---------------------------------------------------------\n" );
		printf( " Command-line options:\n" );
		printf( "   -f    Fullscreen mode\n" );
		printf( "---------------------------------------------------------\n" );
	}
	
	// Initialize everything
	if (0 != DaveStartup(bfullscreen))
	{
		djMSG("main(): initialization failed.");
		return -1;
	}
	
	// Enter main menu
	DoMainMenu();
	
	// Cleanup
	DaveCleanup();
	
	return 0;
}

int DaveStartup(bool bFullScreen)
{
	srand(time(NULL));					// seed the random number generator
	
	djTimeInit();						// Initialise timer

	//-- Initialize graphics
	TRACE("DaveStartup(): Initializing graphics system ...\n");
	//if (!GraphInit( bFullScreen, 640, 480 )) // <-- 640,480 actually "works", sort of
	if (!GraphInit( bFullScreen, 320, 200 ))
	{
		TRACE( "DaveStartup(): Graphics initialization failed.\n" );
		return -1;
	}
	
	djSoundInit();						// Initialize sound

	pImgMain = new djImage;				// Load main skin (title screen)
	pImgMain->Load("data/main.tga");
	
	// Load missions
	if (0 != LoadMissions("data/missions.txt"))
		return -1;
	TRACE( "DaveStartup(): %d missions(s) found.\n", g_apMissions.size() );
	
	TRACE("DaveStartup(): level_init()\n");
	level_init();
	
	//-- Initialize input devices
	TRACE("DaveStartup(): Initializing keys ..\n");
	if (!djiInit( pVisMain, INPUT_KEYDOWN|INPUT_KEYUP|INPUT_KEYREPEAT ))
		return -1;

	mainMenu.m_clrBack = djColor(10,40,150); // Crap colour. Need something better, like a bitmap
	//mainMenu.m_clrBack = djColor(129,60,129);
	mainMenu.m_iSize = 0;
	mainMenu.m_items = mainMenuItems;
	mainMenu.m_szCursor = mainMenuCursor;
	mainMenu.m_xOffset = -1;
	mainMenu.m_yOffset = -1;
	mainMenu.m_iSoundMove = djSoundLoad( "data/sounds/cardflip.wav" );

	LoadHighScores();					// Load high scores

	GameInitialSetup();					// Once-off setup for the game itself (e.g. create in-game menu etc)

	return 0;
}

void DaveCleanup()
{
	GameFinalCleanup();	// Game
	SaveHighScores();	// Save high scores
	djiDone();			// Input
	djSoundDone();		// Sound
	GraphDone();		// Graphics
	djTimeDone();		// Timer stuff
}

void DoMainMenu()
{
	bool brunning = true;
	do
	{
		int menu_option;
		
		// Load main menu background image
		if (pImgMain)
			djgDrawImage( pVisBack, pImgMain, 0, 0, pImgMain->Width(), pImgMain->Height() );
		GraphFlip();

		// Random select menu cursor, either hearts or skulls
		mainMenu.m_szCursor = (rand()%4==0 ? mainMenuCursorSkull : mainMenuCursor);
		
		menu_option = do_menu( &mainMenu );
		
		switch (menu_option)
		{
		case 1: // start new game
			StartNewGame();
			break;
		case 2: // select mission
			SelectMission();
			break;
		case 6: // instructions
			ShowInstructions();
			break;
		case 8:
			ShowHighScores();
			break;
		case 9: // credits
			ShowCredits();
			break;
		case -1: // escape
		case 10: // quit
			brunning = false;
			break;
		}
		
		// is this necessary?
		if (g_iKeys[DJKEY_ESC] == 1)
			brunning = false;
	} while (brunning);
}

void StartNewGame()
{
	int nScore = game_startup();
	if (IsNewHighScore(nScore))
	{
		AddHighScore("Todo, get username", nScore);
		SaveHighScores();
	}
}

void SelectMission()
{
	int i, iret, maxlen;
	SMenuItem *pMenuItems;
	unsigned char menuMissionCursor[] = { 128, 129, 130, 131, 0 };
	CMenu menuMission;
	menuMission.m_clrBack = djColor(129,0,129);
	menuMission.m_iSize = 0;
	menuMission.m_szCursor = menuMissionCursor;
	menuMission.m_xOffset = -1;
	menuMission.m_yOffset = -1;
	menuMission.m_items = NULL;
	
	// Build the menu, adding mission names as entries
	maxlen = 0;
	pMenuItems = new SMenuItem[g_apMissions.size() + 3];
	for ( i=0; i<(int)g_apMissions.size(); i++ )
	{
		pMenuItems[i+1].m_bitem = true;
		pMenuItems[i+1].m_szText = new char[128];
		sprintf( pMenuItems[i+1].m_szText, "|  %-31.31s |", g_apMissions[i]->GetName() );
	}
	// Top and bottom menu entries, the borders
	pMenuItems[0].m_bitem = false;
	pMenuItems[0].m_szText = djStrDeepCopy( "{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}" );
	
	pMenuItems[g_apMissions.size()+1].m_bitem = false;
	pMenuItems[g_apMissions.size()+1].m_szText = djStrDeepCopy("[~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~]");
	pMenuItems[g_apMissions.size()+2].m_bitem = false; // Null-terminator
	pMenuItems[g_apMissions.size()+2].m_szText = NULL;
	
	menuMission.m_items = pMenuItems;
	
	// Do menu
	iret = do_menu( &menuMission );
	if (iret != -1)
	{
		g_pCurMission = g_apMissions[iret - 1];
	}
	
	// Unallocate menu
	for ( i=0; i<(int)g_apMissions.size() + 2; i++ )
	{
		if (pMenuItems[i].m_szText)
		{
			DELV(pMenuItems[i].m_szText);
		}
	}
	DELV(pMenuItems);
}

